using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.IO;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading;
using System.Windows.Forms;

#if false
using Smdn.Collections;
#else
using System.Linq;
#endif
using Smdn.IO;
using Smdn.Formats;
using Smdn.Formats.Ini;
using Smdn.Formats.Mime;
using Smdn.Net.Imap4;
using Smdn.Net.Imap4.Client;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Net;
using Smdn.Net.Pop3;
using Smdn.Net.Pop3.Client;
using Smdn.Net.Pop3.Protocol.Client;
using Smdn.Windows.Forms;

namespace Smdn.Applications.SimpleMailer {
  public partial class MainWindow {
    public MainWindow()
    {
      ImapConnection.ServerCertificateValidationCallback += ValidateServerCertificate;
      PopConnection.ServerCertificateValidationCallback += ValidateServerCertificate;
#if false
      ImapWebRequestDefaults.ClientID["version"] = System.Reflection.Assembly.GetExecutingAssembly().GetName().Version.ToString();
#endif

      InitializeComponent();
      Initialize();
    }

    private static class Fonts {
      public static Font BoldFont;
      public static Font StrikeoutFont;
      public static Font Monospace;

      static Fonts()
      {
        BoldFont = new Font(Control.DefaultFont, FontStyle.Bold);
        StrikeoutFont = new Font(Control.DefaultFont, FontStyle.Strikeout);
        Monospace = new Font(FontFamily.GenericMonospace, Control.DefaultFont.SizeInPoints, FontStyle.Regular, GraphicsUnit.Point);
      }
    }

    private void Initialize()
    {
      InitializeContextMenu();

      toolStripMenuItemFile_Quit.Click += delegate { Close(); };

      toolStripMenuItemMailbox_Update.Click       += delegate(object sender, EventArgs e) { UpdateNode((sender as TreeViewContextMenuItem).ContextNode); };
      toolStripMenuItemMailbox_CancelDelete.Click += delegate(object sender, EventArgs e) { CancelDelete((sender as TreeViewContextMenuItem).ContextNode as PopMailboxNode); };
      toolStripMenuItemMailbox_Append.Click       += delegate(object sender, EventArgs e) { AppendMessage((sender as TreeViewContextMenuItem).ContextNode as ImapMailboxNode); };
      toolStripMenuItemMailbox_Create.Click       += delegate(object sender, EventArgs e) { CreateMailbox((sender as TreeViewContextMenuItem).ContextNode); };
      toolStripMenuItemMailbox_Delete.Click       += delegate(object sender, EventArgs e) { DeleteMailbox((sender as TreeViewContextMenuItem).ContextNode as ImapMailboxNode); };
      toolStripMenuItemMailbox_MoveToTrash.Click  += delegate(object sender, EventArgs e) { MoveMailboxToTrash((sender as TreeViewContextMenuItem).ContextNode as ImapMailboxNode); };
      toolStripMenuItemMailbox_Rename.Click       += delegate(object sender, EventArgs e) { RenameMailbox((sender as TreeViewContextMenuItem).ContextNode as ImapMailboxNode); };
      toolStripMenuItemMailbox_Subscribe.Click    += delegate(object sender, EventArgs e) { ShowSubscribeDialog((sender as TreeViewContextMenuItem).ContextNode); };
      toolStripMenuItemMailbox_Unsubscribe.Click  += delegate(object sender, EventArgs e) { UnsubscribeMailbox((sender as TreeViewContextMenuItem).ContextNode as ImapMailboxNode); };

      toolStripMenuItemMessage_Fetch.Click        += delegate(object sender, EventArgs e) { FetchAndSaveMessage((sender as ListViewContextMenuItem).ContextMessage); };
      toolStripMenuItemMessage_ToggleSeen.Click   += delegate(object sender, EventArgs e) { ToggleMessageSeen((sender as ListViewContextMenuItem).ContextMessage as ImapMessageItem); };
      toolStripMenuItemMessage_Expunge.Click      += delegate(object sender, EventArgs e) { ExpungeMessage((sender as ListViewContextMenuItem).ContextMessage); };
      toolStripMenuItemMessage_MoveToTrash.Click  += delegate(object sender, EventArgs e) { MoveMessageToTrash((sender as ListViewContextMenuItem).ContextMessage as ImapMessageItem); };
      toolStripMenuItemMessage_ToggleKeywords_Append.Click
                                                  += delegate(object sender, EventArgs e) { StoreMessageKeyword((sender as ListViewContextMenuItem).ContextMessage as ImapMessageItem); };

      toolStripMenuItemMessage_CopyTo.DropDownOpening += CopyOrMoveMessageToDropDownOpening;
      toolStripMenuItemMessage_MoveTo.DropDownOpening += CopyOrMoveMessageToDropDownOpening;

      toolStripMenuItemMessage_ToggleKeywords.DropDownOpening += ToggleMessageKeywordsDropDownOpening;

      dropDownListSections.SelectedIndexChanged += delegate(object sender, EventArgs e) {
        SectionSelected((sender as SectionDropDownList).SelectedItem);
      };
      buttonFetchAndSaveSection.Click += delegate {
        FetchAndSaveSection(dropDownListSections.SelectedItem);
      };

      treeViewMailbox.NodeMouseClick += AccountOrMailboxSelected;

      listViewMessages.View = View.Details;
      listViewMessages.FullRowSelect = true;
      listViewMessages.HeaderStyle = ColumnHeaderStyle.Clickable;
      listViewMessages.Columns.Add("タイトル", 160);
      listViewMessages.Columns.Add("UID", 40);
      listViewMessages.Columns.Add("サイズ", 60);
      listViewMessages.Columns.Add("Content-Type", 80);
      listViewMessages.Columns.Add("受信日 (internal date)", 120);
      listViewMessages.Columns.Add("送信日", 120);
      listViewMessages.Columns.Add("送信者", 100);
      listViewMessages.Columns.Add("フラグ", 200);
      listViewMessages.SelectedIndexChanged += MessageSelected;
      listViewMessages.ColumnClick += delegate(object sender, ColumnClickEventArgs e) {
        var listView = sender as ListView;
        var sorter = new MessageListViewItemComparer(listView.ListViewItemSorter as MessageListViewItemComparer, e.Column);

        listView.ListViewItemSorter = sorter;
      };

      textBoxMessage.Multiline = true;
      textBoxMessage.ScrollBars = ScrollBars.Both;
      textBoxMessage.Font = Fonts.Monospace;

      textBoxStatus.ReadOnly = true;
      textBoxStatus.Multiline = true;
      textBoxStatus.ScrollBars = ScrollBars.Both;
      textBoxStatus.Font = Fonts.Monospace;

      textBoxServerInfo.ReadOnly = true;
      textBoxServerInfo.Multiline = true;
      textBoxServerInfo.ScrollBars = ScrollBars.Both;
      textBoxServerInfo.Font = Fonts.Monospace;

      timerUpdate.Interval = 5 * 60 * 1000; // 5 min
      timerUpdate.Tick += delegate {
        NoOp();
      };
      timerUpdate.Enabled = true;
    }

    private ToolStripItem[] mailboxToolStripItems;
    private ToolStripItem[] messageToolStripItems;

    private void SetMailboxContextMenuProperty(Point? cursorPosition)
    {
      var contextNode = cursorPosition.HasValue
        ? treeViewMailbox.GetNodeAt(treeViewMailbox.PointToClient(cursorPosition.Value))
        : treeViewMailbox.SelectedNode;

      var isImap = (contextNode is ImapMailboxNode) || (contextNode is ImapAccountNode);
      var mailboxSelected = contextNode is ImapMailboxNode;
      var mailbox = mailboxSelected ? (contextNode as ImapMailboxNode).Mailbox : null;

      toolStripMenuItemMailbox_Append.Enabled =
      toolStripMenuItemMailbox_MoveToTrash.Enabled =
      toolStripMenuItemMailbox_Delete.Enabled =
      toolStripMenuItemMailbox_Rename.Enabled =
      toolStripMenuItemMailbox_Unsubscribe.Enabled = mailboxSelected;

      toolStripMenuItemMailbox_Append.Visible =
      toolStripMenuItemMailbox_Create.Visible =
      toolStripMenuItemMailbox_MoveToTrash.Visible =
      toolStripMenuItemMailbox_Delete.Visible =
      toolStripMenuItemMailbox_Rename.Visible =
      toolStripMenuItemMailbox_Subscribe.Visible =
      toolStripMenuItemMailbox_Unsubscribe.Visible = isImap;

      toolStripMenuItemMailbox_CancelDelete.Visible = !isImap && contextNode is IMailboxContainer;

      toolStripMenuItemMailbox_Append.ContextNode =
      toolStripMenuItemMailbox_MoveToTrash.ContextNode =
      toolStripMenuItemMailbox_Delete.ContextNode =
      toolStripMenuItemMailbox_Rename.ContextNode =
      toolStripMenuItemMailbox_Unsubscribe.ContextNode = contextNode;

      var accountSelected = contextNode is ImapAccountNode;

      toolStripMenuItemMailbox_CancelDelete.Enabled =
      toolStripMenuItemMailbox_Update.Enabled = contextNode is IAccountContainer ||
                                                contextNode is IMailboxContainer;

      toolStripMenuItemMailbox_Create.Enabled =
      toolStripMenuItemMailbox_Subscribe.Enabled = mailboxSelected || accountSelected;

      toolStripMenuItemMailbox_Create.Enabled &= (mailbox == null || mailbox.CanHaveChild);

      toolStripMenuItemMailbox_CancelDelete.ContextNode =
      toolStripMenuItemMailbox_Update.ContextNode =
      toolStripMenuItemMailbox_Create.ContextNode =
      toolStripMenuItemMailbox_Subscribe.ContextNode = contextNode;
    }

    private void SetMessageContextMenuProperty()
    {
      var contextMessage     = (0 < listViewMessages.SelectedItems.Count) ? listViewMessages.SelectedItems[0] as IMessageContainer : null;
      var isImap = contextMessage is ImapMessageItem;
      var contextImapMessage = contextMessage as ImapMessageItem;

      toolStripMenuItemMessage_Fetch.Enabled = 
      toolStripMenuItemMessage_Expunge.Enabled = (contextMessage != null && !contextMessage.IsDeleted);

      toolStripMenuItemMessage_MoveTo.Enabled = 
      toolStripMenuItemMessage_CopyTo.Enabled = 
      toolStripMenuItemMessage_ToggleSeen.Enabled = 
      toolStripMenuItemMessage_MoveToTrash.Enabled =
      toolStripMenuItemMessage_ToggleKeywords.Enabled = (contextImapMessage != null && !contextMessage.IsDeleted);

      toolStripMenuItemMessage_MoveTo.Visible = 
      toolStripMenuItemMessage_CopyTo.Visible = 
      toolStripMenuItemMessage_ToggleSeen.Visible = 
      toolStripMenuItemMessage_MoveToTrash.Visible = 
      toolStripMenuItemMessage_ToggleKeywords.Visible = isImap;

      toolStripMenuItemMessage_Fetch.ContextMessage = 
      toolStripMenuItemMessage_MoveTo.ContextMessage = 
      toolStripMenuItemMessage_CopyTo.ContextMessage = 
      toolStripMenuItemMessage_ToggleSeen.ContextMessage = 
      toolStripMenuItemMessage_Expunge.ContextMessage = 
      toolStripMenuItemMessage_MoveToTrash.ContextMessage = 
      toolStripMenuItemMessage_ToggleKeywords.ContextMessage = contextMessage;

      toolStripMenuItemMessage_MoveTo.DropDownItems.Clear();
      toolStripMenuItemMessage_MoveTo.DropDownItems.Add(contextMenuDummyItem);

      toolStripMenuItemMessage_CopyTo.DropDownItems.Clear();
      toolStripMenuItemMessage_CopyTo.DropDownItems.Add(contextMenuDummyItem);

      if (contextImapMessage == null)
        toolStripMenuItemMessage_ToggleSeen.Checked = false;
      else
        toolStripMenuItemMessage_ToggleSeen.Checked = contextImapMessage.Message.Flags.Contains(ImapMessageFlag.Seen);
    }

    private void InitializeContextMenu()
    {
      mailboxToolStripItems = new ToolStripItem[toolStripMenuItemMailbox.DropDownItems.Count];

      toolStripMenuItemMailbox.DropDownItems.CopyTo(mailboxToolStripItems, 0);

      messageToolStripItems = new ToolStripItem[toolStripMenuItemMessage.DropDownItems.Count];

      toolStripMenuItemMessage.DropDownItems.CopyTo(messageToolStripItems, 0);

      contextMenuMailbox.Items.Add(contextMenuDummyItem);

      contextMenuMailbox.Opening += delegate {
        SetMailboxContextMenuProperty(Cursor.Position);

        toolStripMenuItemMailbox.DropDownItems.Clear();
        contextMenuMailbox.Items.Clear();
        contextMenuMailbox.Items.AddRange(mailboxToolStripItems);
      };

      contextMenuMailbox.Closed += delegate {
        contextMenuMailbox.Items.Clear();
        contextMenuMailbox.Items.Add(contextMenuDummyItem);
        toolStripMenuItemMailbox.DropDownItems.Clear();
        toolStripMenuItemMailbox.DropDownItems.AddRange(mailboxToolStripItems);
      };

      toolStripMenuItemMailbox.DropDownOpening += delegate {
        SetMailboxContextMenuProperty(null);

        contextMenuMailbox.Items.Clear();
        contextMenuMailbox.Items.Add(contextMenuDummyItem);
        toolStripMenuItemMailbox.DropDownItems.Clear();
        toolStripMenuItemMailbox.DropDownItems.AddRange(mailboxToolStripItems);
      };

      toolStripMenuItemMailbox.DropDownClosed += delegate {
        toolStripMenuItemMailbox.DropDownItems.Clear();
        contextMenuMailbox.Items.Clear();
        contextMenuMailbox.Items.AddRange(mailboxToolStripItems);
      };


      contextMenuMessage.Items.Add(contextMenuDummyItem);

      contextMenuMessage.Opening += delegate {
        SetMessageContextMenuProperty();

        toolStripMenuItemMessage.DropDownItems.Clear();
        contextMenuMessage.Items.Clear();
        contextMenuMessage.Items.AddRange(messageToolStripItems);
      };

      contextMenuMessage.Closed += delegate {
        contextMenuMessage.Items.Clear();
        contextMenuMessage.Items.Add(contextMenuDummyItem);
        toolStripMenuItemMessage.DropDownItems.Clear();
        toolStripMenuItemMessage.DropDownItems.AddRange(messageToolStripItems);
      };

      toolStripMenuItemMessage.DropDownOpening += delegate {
        SetMessageContextMenuProperty();

        contextMenuMessage.Items.Clear();
        contextMenuMessage.Items.Add(contextMenuDummyItem);
        toolStripMenuItemMessage.DropDownItems.Clear();
        toolStripMenuItemMessage.DropDownItems.AddRange(messageToolStripItems);
      };

      toolStripMenuItemMessage.DropDownClosed += delegate {
        toolStripMenuItemMessage.DropDownItems.Clear();
        contextMenuMessage.Items.Clear();
        contextMenuMessage.Items.AddRange(messageToolStripItems);
      };
    }

    protected override void OnShown (EventArgs e)
    {
      base.OnShown (e);

      Application.DoEvents();

      LoadConfig();

      InitializeMailboxTree();
    }

    protected override void OnClosing (CancelEventArgs e)
    {
      base.OnClosing (e);

      if (!e.Cancel) {
        LogoutAll();
      }
    }

    private bool ValidateServerCertificate(object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
    {
      if (sslPolicyErrors == SslPolicyErrors.None)
        return true;
      else if (!verifyServerCertificates)
        return true;

      var message = new StringBuilder();

      message.AppendLine("サーバ証明書に問題があります。　続行しますか?");
      message.AppendLine();
      message.AppendFormat("SslPolicyErrors: {0}", sslPolicyErrors);
      message.AppendLine();
      message.AppendFormat("X509Certificate: {0}{1}", Environment.NewLine, certificate.ToString(false));
      message.AppendLine();

      if (chain != null) {
        foreach (var element in chain.ChainElements) {
          message.AppendFormat("発行先: {0}", element.Certificate.Subject);
          message.AppendLine();
          message.AppendFormat("発行者: {0}", element.Certificate.Issuer);
          message.AppendLine();
          message.AppendFormat("有効期間: {0} から {1}", element.Certificate.GetEffectiveDateString(), element.Certificate.GetExpirationDateString());
          message.AppendLine();

          foreach (var status in element.ChainElementStatus) {
            message.AppendFormat("Status: {0}", status.Status);
            message.AppendLine();
            message.AppendFormat("StatusInformation: {0}", status.StatusInformation);
            message.AppendLine();
            message.AppendLine();
          }
        }
      }

      var result = MessageBox.Show(message.ToString(),
                                   "ValidateServerCertificate",
                                   MessageBoxButtons.YesNo);

      return (result == DialogResult.Yes);
    }

    private void LoadConfig()
    {
      Converter<string, bool> convertYesNo = (val) => val != "no";
      var config = IniDocument.Load("SimpleMailer.ini");
      var accounts = new List<AccountBase>();

      var defaultSection = config.DefaultSection;

      mimeTypesFile = defaultSection.Get("MimeTypesFile", "/etc/mime.types");
      verifyServerCertificates = defaultSection.Get("VerifyServerCertificates", true, convertYesNo);

      // 証明書ファイルのセクション
      var certFilesSection = config["CertFiles"];

      foreach(var entry in certFilesSection) {
        try {
          ImapConnection.ClientCertificates.Add(new X509Certificate2(entry.Value));
          PopConnection.ClientCertificates.Add(new X509Certificate2(entry.Value));
        }
        catch (Exception ex) {
          AppendStatus("証明書{0}を読み込めませんでした", entry.Value);
          AppendStatus("Exception: {0}", ex);
        }
      }

      config.Remove(certFilesSection.Name);

      // アカウントのセクション
      foreach (var section in config) {
        if (section.IsDefaultSection)
          continue;

        try {
          var serverUri = section.GetThrowException<Uri>("ServerURL", ConvertUtils.ToUri);

          AccountBase account;

          if (ImapUri.IsImap(serverUri)) {
            var username = ImapStyleUriParser.GetUser(serverUri);
            var credential = section.Entries.ContainsKey("Password")
              ? (ICredentialsByHost)new NetworkCredential(username, section["Password"])
              : (ICredentialsByHost)new PasswordInputCredential(username, this);

            account = new ImapAccount(section.Name,
                                      serverUri,
                                      credential,
                                      section.Get("Timeout", System.Threading.Timeout.Infinite, Convert.ToInt32),
                                      section.Get("TryTLS", true, convertYesNo),
                                      section.Get("ReadOnly", true, convertYesNo));
          }
          else if (PopUri.IsPop(serverUri)) {
            var username = PopStyleUriParser.GetUser(serverUri);
            var credential = section.Entries.ContainsKey("Password")
              ? (ICredentialsByHost)new NetworkCredential(username, section["Password"])
              : (ICredentialsByHost)new PasswordInputCredential(username, this);

            account = new PopAccount(section.Name,
                                     serverUri,
                                     credential,
                                     section.Get("Timeout", System.Threading.Timeout.Infinite, Convert.ToInt32),
                                     section.Get("TryTLS", true, convertYesNo));
          }
          else {
            throw new UriFormatException("unsupported uri scheme");
          }

          accounts.Add(account);
        }
        catch (Exception ex) {
          AppendStatus("{0}のアカウント設定を読み込めませんでした", section.Name);
          AppendStatus("Exception: {0}", ex);
        }
      }

      this.imapAccounts = accounts.FindAll(delegate(AccountBase account) { return account is ImapAccount; })
                                  .ConvertAll(delegate(AccountBase account) { return account as ImapAccount; }).ToArray();
      this.popAccounts = accounts.FindAll(delegate(AccountBase account) { return account is PopAccount; })
                                 .ConvertAll(delegate(AccountBase account) { return account as PopAccount; }).ToArray();
      this.allAccounts = accounts.ToArray();

      if (this.allAccounts.Length == 0)
        AppendStatus("表示できるアカウント設定がありません。");
    }

    private void InitializeMailboxTree()
    {
      foreach (var account in imapAccounts) {
        UpdateMailboxTree(account);
      }

      foreach (var account in popAccounts) {
        UpdateMailboxTree(account);
      }
    }

    private void UpdateMailboxTree(object sender, EventArgs e)
    {
      if (sender is TreeViewContextMenuItem) {
        var node = (sender as TreeViewContextMenuItem).ContextNode;

        if (node is IAccountContainer)
          UpdateMailboxTree((node as IAccountContainer).Account);
      }
    }

    private void UpdateMailboxTree(AccountBase account)
    {
      if (account is ImapAccount)
        UpdateMailboxTree(account as ImapAccount);
      else if (account is PopAccount)
        UpdateMailboxTree(account as PopAccount);
    }

    private void UpdateMailboxTree(ImapAccount account)
    {
      try {
        account.EnsureConnected();
      }
      catch (Exception ex) {
        AppendStatus("{0}の接続に失敗しました", account.Name);
        AppendStatus(ex);
        return;
      }

      account.Client.ExistMessageCountChanged   += MailboxSizeChanged;
      account.Client.RecentMessageCountChanged  += MailboxSizeChanged;

      account.Client.MessageDeleted       += MessageStatusChanged;
      account.Client.MessageStatusChanged += MessageStatusChanged;

      // アカウントとメールボックスのノードを作成する
      try {
        Cursor = Cursors.WaitCursor;

        treeViewMailbox.BeginUpdate();

        var removingNodes = new List<ImapMailboxNode>();

        foreach (var pair in imapMailboxNodeMap) {
          if (pair.Value.Account == account)
            removingNodes.Add(pair.Value);
        }

        foreach (var removingNode in removingNodes) {
          imapMailboxNodeMap.Remove(removingNode.Mailbox);
        }

        ImapAccountNode rootNode = null;

        foreach (TreeNode node in treeViewMailbox.Nodes) {
          if ((node is ImapAccountNode) && (node as ImapAccountNode).Account == account) {
            rootNode = node as ImapAccountNode;
            break;
          }
        }

        if (rootNode == null) {
          rootNode = new ImapAccountNode(account);
          treeViewMailbox.Nodes.Add(rootNode);
        }
        else {
          rootNode.Nodes.Clear();
        }

        try {
          const ImapMailboxListOptions options =
            ImapMailboxListOptions.SubscribedOnly | ImapMailboxListOptions.RequestStatus;

          var sortedMailboxes = new List<ImapMailboxInfo>(account.Client.GetMailboxes(options));

          sortedMailboxes.Sort(delegate(ImapMailboxInfo x, ImapMailboxInfo y) {
            return string.Compare(x.FullName, y.FullName);
          });

          foreach (var mailbox in sortedMailboxes) {
            var node = new ImapMailboxNode(account, mailbox);

            if (mailbox.FullName.Contains(mailbox.MailboxSeparator)) {
              // find parent
              var hierarchy = mailbox.FullName.Split(new [] {mailbox.MailboxSeparator}, StringSplitOptions.None);
              TreeNode parent = rootNode;
              var found = false;

              for (var i = 0; i < hierarchy.Length - 1; i++) {
                found = false;

                foreach (TreeNode n in parent.Nodes) {
                  if (n is ImapMailboxNode) {
                    if ((n as ImapMailboxNode).Mailbox.FullName == hierarchy[i + 1]) {
                      found = true;
                      parent = n;
                      break;
                    }
                  }
                }

                if (!found)
                  break;
              }

              if (!found) {
                parent = rootNode;

                for (var i = 0; i < hierarchy.Length - 1; i++) {
                  var name = string.Join(mailbox.MailboxSeparator, hierarchy, 0, i + 1);
                  var n = rootNode.Nodes.Find(name, true).FirstOrDefault();

                  if (n == null) {
                    n = new ImapNonExistentMailboxNode(hierarchy[i], name);
                    parent.Nodes.Add(n);
                    parent = n;
                  }
                  else {
                    parent = n;
                  }
                }
              }

              parent.Nodes.Add(node);
            }
            else {
              rootNode.Nodes.Add(node);
            }

            imapMailboxNodeMap.Add(mailbox, node);
          }

        }
        catch (Exception ex) {
          AppendStatus("{0}のメールボックス一覧を取得できませんでした", account.Name);
          AppendStatus(ex);
          return;
        }

        rootNode.ExpandAll();
      }
      finally {
        treeViewMailbox.EndUpdate();

        Cursor = Cursors.Default;
      }

      treeViewMailbox.Refresh();
    }

    private void UpdateMailboxTree(PopAccount account)
    {
      try {
        account.EnsureConnected();
      }
      catch (Exception ex) {
        AppendStatus("{0}の接続に失敗しました", account.Name);
        AppendStatus(ex);
        return;
      }

      // アカウントのノードを作成する
      PopAccountNode rootNode = null;

      try {
        Cursor = Cursors.WaitCursor;

        treeViewMailbox.BeginUpdate();

        foreach (TreeNode node in treeViewMailbox.Nodes) {
          if ((node is PopAccountNode) && (node as PopAccountNode).Account == account) {
            rootNode = node as PopAccountNode;
            break;
          }
        }

        if (rootNode == null) {
          rootNode = new PopAccountNode(account);
          treeViewMailbox.Nodes.Add(rootNode);
        }
        else {
          rootNode.Nodes.Clear();
        }

        var mailboxNode = new PopMailboxNode(account);

        rootNode.Nodes.Add(mailboxNode);

        popMailboxNodeMap[account] = mailboxNode;

        rootNode.ExpandAll();
      }
      finally {
        treeViewMailbox.EndUpdate();

        Cursor = Cursors.Default;
      }

      treeViewMailbox.Refresh();
    }

    private void UpdateNode(TreeNode node)
    {
      if (node is ImapMailboxNode) {
        UpdateMailbox(node as ImapMailboxNode);
      }
      else if (node is ImapAccountNode) {
        UpdateMailboxTree((node as ImapAccountNode).Account);
      }
      else if (node is PopMailboxNode) {
        UpdateMailboxTree((node as PopMailboxNode).Account);
      }
      else if (node is PopAccountNode) {
        UpdateMailboxTree((node as PopAccountNode).Account);
      }
    }

    private void UpdateMailbox(ImapMailboxNode node)
    {
      try {
        // XXX
        if (node.Mailbox.IsOpen) {
          node.Account.Client.OpenedMailbox.Refresh();
        }
        else {
          node.Mailbox.Refresh();
          node.UpdateText();
        }
      }
      catch (Exception ex) {
        AppendStatus("{0}を最新の状態に更新できませんでした", node.Mailbox.Name);
        AppendStatus(ex);
      }
    }

    private void MailboxSizeChanged(object sender, ImapMailboxSizeChangedEventArgs e)
    {
      foreach (var pair in imapMailboxNodeMap) {
        if (pair.Key.Client != sender)
          continue;

        if (pair.Key.FullName == e.Mailbox.FullName) {
          pair.Value.UpdateText();
          break;
        }
      }
    }

    private void MessageStatusChanged(object sender, ImapMessageStatusChangedEventArgs e)
    {
      var changedMessages = new List<ImapMessageInfo>(e.Messages);

      foreach (var item in listViewMessages.Items) {
        var i = item as ImapMessageItem;

        if (i == null)
          continue;

        if (changedMessages.Exists(delegate(ImapMessageInfo m) { return m == i.Message; })) {
          i.UpdateView();

          changedMessages.Remove(i.Message);

          if (changedMessages.Count == 0)
            return;
        }
      }
    }

    private void AccountOrMailboxSelected(object sender, TreeNodeMouseClickEventArgs e)
    {
      if ((e.Button & MouseButtons.Left) != MouseButtons.Left)
        return;

      if (e.Node is IMailboxContainer) {
        var mailboxContainer = e.Node as IMailboxContainer;

        ViewServerInfo(mailboxContainer);

        UpdateMessageList(mailboxContainer);

        UpdateQuotaUsage(mailboxContainer);

        AppendStatus(mailboxContainer.GetStatusText());
      }
      else if (e.Node is IAccountContainer) {
        ViewServerInfo(e.Node as IAccountContainer);
      }
    }

    private void UpdateMessageList()
    {
      var mailboxContainer = treeViewMailbox.SelectedNode as IMailboxContainer;

      if (mailboxContainer == null)
        return;
      else
        UpdateMessageList(mailboxContainer);
    }

    private void UpdateMessageList(IMailboxContainer container)
    {
      if (container.Account is ImapAccount)
        UpdateMessageList(container.Account as ImapAccount, container.GetMailbox() as ImapMailboxInfo);
      else
        UpdateMessageList(container.Account as PopAccount);

      dropDownListSections.Items.Clear();
      dropDownListSections.Enabled = false;
    }

    private void UpdateMessageList(ImapAccount account, ImapMailboxInfo mailbox)
    {
      if (!mailbox.Exists || mailbox.IsUnselectable) {
        AppendStatus("{0}は選択できません", mailbox.Name);
        return;
      }
      else if (mailbox == account.Client.OpenedMailbox) {
        return;
      }

      var mailboxNode = imapMailboxNodeMap[mailbox];

      // メッセージの一覧を取得
      AppendStatus("{0}のメッセージの一覧を取得しています", mailbox.Name);

      try {
        if (account.Client.OpenedMailbox != null)
          account.Client.OpenedMailbox.Close();

        account.Client.OpenMailbox(mailbox, account.ReadOnly);

        mailboxNode.UpdateText();

        toolStripProgressBarProgress.Minimum = 0;
        toolStripProgressBarProgress.Maximum = (int)account.Client.OpenedMailbox.ExistMessageCount;
        toolStripProgressBarProgress.Value = 0;
        toolStripProgressBarProgress.Step = 1;
        toolStripProgressBarProgress.Style = ProgressBarStyle.Blocks;

        CreateMessageList(account, account.Client.OpenedMailbox.GetMessages(ImapMessageFetchAttributeOptions.AllAttributes));

        AppendStatus("全{0}件のメッセージがあります",account.Client.OpenedMailbox.ExistMessageCount);
      }
      catch (Exception ex) {
        AppendStatus("{0}のメッセージ一覧を取得できませんでした", mailbox.Name);
        AppendStatus(ex);
      }
    }

    private void CreateMessageList(ImapAccount account, IEnumerable<ImapMessageInfo> messages)
    {
      try {
        listViewMessages.BeginUpdate();
        listViewMessages.Items.Clear();

        foreach (var message in messages) {
          listViewMessages.Items.Add(new ImapMessageItem(account, message));

          toolStripProgressBarProgress.Increment(1);
        }
      }
      finally {
        listViewMessages.EndUpdate();
      }
    }

    private void UpdateMessageList(PopAccount account)
    {
      // メッセージの一覧を取得
      AppendStatus("{0}のメッセージの一覧を取得しています", account.Name);

      try {
        var messageCount = 0;

        CreateMessageList(account, account.Client.GetMessages(true).Select(delegate(PopMessageInfo info){
          messageCount++;

          return Tuple.Create(info, info.ReadAs<MimeMessage>(0, MimeMessage.Load));
        }));

        AppendStatus("全{0}件のメッセージがあります", messageCount);
      }
      catch (Exception ex) {
        AppendStatus("{0}のメッセージ一覧を取得できませんでした", account.Name);
        AppendStatus(ex);
      }
    }

    private void CreateMessageList(PopAccount account, IEnumerable<Tuple<PopMessageInfo, MimeMessage>> messages)
    {
      try {
        listViewMessages.BeginUpdate();
        listViewMessages.Items.Clear();

        foreach (var message in messages) {
          listViewMessages.Items.Add(new PopMessageItem(account, message.Item1, message.Item2));
        }
      }
      finally {
        listViewMessages.EndUpdate();
      }
    }

    private void UpdateQuotaUsage(IMailboxContainer container)
    {
      var mailbox = container.GetMailbox() as ImapMailboxInfo;

      if (mailbox == null)
        return;

      var removingItems = new List<QuotaResourceToolStripItem>();

      foreach (var item in statusStripMain.Items) {
        var quotaResourceItem = item as QuotaResourceToolStripItem;

        if (quotaResourceItem != null)
          removingItems.Add(quotaResourceItem);
      }

      foreach (var item in removingItems) {
        statusStripMain.Items.Remove(item);
      }

      if (!mailbox.Exists)
        return;

      AppendStatus("{0}のクォータ使用率を取得しています", mailbox.Name);

      try {
        foreach (var quota in mailbox.GetQuota()) {
          foreach (var resource in quota.Resources) {
            var item = new QuotaResourceToolStripItem();

            item.Resource= resource;
            item.QuotaRoot = quota.Root;

            statusStripMain.Items.Add(item);
          }
        }

        AppendStatus("{0}のクォータ使用率を取得しました", mailbox.Name);
      }
      catch (Exception ex) {
        AppendStatus("{0}のクォータ使用率を取得できませんでした", mailbox.Name);
        AppendStatus(ex);
      }
    }

    private void MessageSelected(object sender, EventArgs e)
    {
      var container = (listViewMessages.SelectedItems.Count <= 0)
        ? null
        : listViewMessages.SelectedItems[0] as IMessageContainer;

      if (container is ImapMessageItem)
        UpdateSectionsDropDownList(container.Account, (container as ImapMessageItem).Message);
      else
        UpdateSectionsDropDownList(null, null);

      if (container == null)
        return;

      MessageSelected(container);
    }

    private void MessageSelected(IMessageContainer container)
    {
      var sw = new Stopwatch();
      MimeMessage message = null;

      if (container.IsDeleted) {
        AppendStatus("{0}は削除されています", container.UniqueId);
        return;
      }

      AppendStatus("{0}の本文を取得しています", container.UniqueId);

      try {
        sw.Start();

        try {
          //message = container.ReadAs<MimeMessage>(MimeMessage.Load);
          using (var stream = container.OpenRead(toolStripProgressBarProgress)) {
            message = MimeMessage.Load(stream);
          }
        }
        catch (Exception ex) {
          AppendStatus("本文の解析に失敗しました");
          AppendStatus("Exception: {0}", ex);
        }

        sw.Stop();
      }
      catch (Exception ex) {
        AppendStatus("{0}の本文を取得できませんでした", container.UniqueId);
        AppendStatus(ex);
        return;
      }

      if (message != null)
        ViewMessage(message);

      AppendStatus("{0}の本文を取得しました ({1:N2}secs)", container.UniqueId, sw.ElapsedMilliseconds / 1000.0);
    }

    private void UpdateSectionsDropDownList(AccountBase account, ImapMessageInfo message)
    {
      var bodyStructure = (message == null) ? null : message.BodyStructure;

      dropDownListSections.Enabled = (bodyStructure != null);

      dropDownListSections.BeginUpdate();

      try {
        dropDownListSections.Items.Clear();

        if (bodyStructure == null) {
          dropDownListSections.SelectedIndex = -1;
        }
        else {
          dropDownListSections.Items.Add(new SectionDropDownListItem(account, message, bodyStructure));

          bodyStructure.Traverse(delegate(IImapBodyStructure nestedSection) {
            dropDownListSections.Items.Add(new SectionDropDownListItem(account, message, nestedSection));
          });

          try {
            sectionSelectionLock = true;

            dropDownListSections.SelectedIndex = 0; // this will cause of calling SectionSelected
          }
          finally {
            sectionSelectionLock = false;
          }
        }
      }
      finally {
        dropDownListSections.EndUpdate();
      }
    }

    private bool sectionSelectionLock = false;

    private void SectionSelected(SectionDropDownListItem sectionItem)
    {
      if (sectionItem.IsDeleted) {
        AppendStatus("{0}は削除されています", sectionItem.UniqueId);
        return;
      }

      if (sectionSelectionLock)
        return;

      if (sectionItem.Structure.MediaType.TypeEqualsIgnoreCase("multipart") ||
          sectionItem.Structure.MediaType.TypeEqualsIgnoreCase("message")) {
        MessageSelected(sectionItem);
      }
      else if (sectionItem.Structure.MediaType.TypeEqualsIgnoreCase("text")) {
        string message = null;
        var sw = new Stopwatch();

        AppendStatus("{0}の本文を取得しています", sectionItem.UniqueId);

        try {
          sw.Start();

          try {
            message = sectionItem.Message.ReadAllText(sectionItem.Structure, ImapMessageFetchBodyOptions.DecodeContent);
          }
          catch (Exception ex) {
            AppendStatus("本文のデコードに失敗しました");
            AppendStatus("Exception: {0}", ex);
          }

          sw.Stop();
        }
        catch (Exception ex) {
          AppendStatus("{0}の本文を取得できませんでした", sectionItem.UniqueId);
          AppendStatus(ex);
          return;
        }

        if (message != null)
          ViewMessage(message);

        AppendStatus("完了 ({0:N2}secs)", sw.ElapsedMilliseconds / 1000.0);
      }
      else {
        var sb = new StringBuilder(4096);

        sb.AppendLine("(非テキストメッセージです)");
        sb.AppendLine();

        var part = sectionItem.Structure as ImapSinglePartBodyStructure;

        if (part == null)
          // XXX
          return;

        sb.AppendFormat("Content-Type: {0}", part.MediaType);
        sb.AppendLine();
        sb.AppendFormat("Size: {0}", part.Size);
        sb.AppendLine();
        sb.AppendFormat("Encoding: {0}", part.Encoding);
        sb.AppendLine();

        var partExt = part as IImapBodyStructureExtension;

        if (partExt != null) {
          if (partExt.Disposition != null) {
            sb.AppendFormat("Disposition: {0} (filename: {1})",
                            partExt.Disposition.Type,
                            MimeEncoding.DecodeNullable(partExt.Disposition.Filename));
            sb.AppendLine();
          }
        }

        ViewMessage(sb.ToString());
      }
    }

    private void ViewMessage(string message)
    {
      textBoxMessage.Text = message;
    }

    private void ViewMessage(MimeMessage message)
    {
      var sb = new StringBuilder(4096);

      if (message.MimeType.TypeEqualsIgnoreCase("text"))
        sb.AppendLine(message.ReadContentAsText());
      else if (message.MimeType.TypeEqualsIgnoreCase("multipart"))
        sb.AppendLine("(マルチパートメッセージです)");
      else
        sb.AppendLine("(非テキストメッセージです)");

      if (0 < message.SubParts.Count) {
        sb.AppendLine("-- multipart message --");
        sb.AppendLine();

        foreach (var subPart in message.SubParts) {
          if (subPart.MimeType.TypeEqualsIgnoreCase("text"))
            sb.AppendLine(subPart.ReadContentAsText());
          else if (subPart.MimeType.TypeEqualsIgnoreCase("multipart"))
            sb.AppendLine("(マルチパートメッセージです)");
          else
            sb.AppendLine("(非テキストメッセージです)");

          sb.AppendLine("-- end of part --");
        }
      }

      textBoxMessage.Text = sb.ToString();
    }

    private void ViewServerInfo(IAccountContainer container)
    {
      if (container.Account is ImapAccount)
        ViewServerInfo(container.Account as ImapAccount);
      else if (container.Account is PopAccount)
        ViewServerInfo(container.Account as PopAccount);
    }

    private void ViewServerInfo(ImapAccount account)
    {
      var sb = new StringBuilder();

      sb.AppendLine("[server capabilities]");
      sb.AppendLine(string.Join(", ", account.Client.ServerCapabilities.ToStringArray()));
      sb.AppendLine();

      sb.AppendLine("[server ID]");

      foreach (var pair in account.Client.ServerID) {
        sb.AppendFormat("{0}: {1}", pair.Key, pair.Value);
        sb.AppendLine();
      }
      sb.AppendLine();

      var namespaces = account.Client.ServerNamespace;

      foreach (var ns in new[] {
        new {Namespaces = namespaces.PersonalNamespaces, Title = "personal"},
        new {Namespaces = namespaces.SharedNamespaces, Title = "shared"},
        new {Namespaces = namespaces.OtherUsersNamespaces, Title = "other users"},
      }) {
        sb.AppendFormat("[namespaces ({0})]", ns.Title);
        sb.AppendLine();
        foreach (var desc in ns.Namespaces) {
          sb.AppendFormat("  prefix: \"{0}\" delimiter: \"{1}\"", desc.Prefix, desc.HierarchyDelimiter);
          sb.AppendLine();
        }
        sb.AppendLine();
      }

      textBoxServerInfo.Text = sb.ToString();
    }

    private void ViewServerInfo(PopAccount account)
    {
      var sb = new StringBuilder();

      sb.AppendLine("[server capabilities]");
      sb.AppendLine(string.Join(", ", account.Client.ServerCapabilities.ToStringArray()));
      sb.AppendLine();

      textBoxServerInfo.Text = sb.ToString();
    }

    private readonly string fileDialogMessageFilterString = FileDialogFilter.CreateFilterString(new[] {
      new[] {"emlファイル", "*.eml"},
      new[] {"すべてのファイル", "*.*"},
    });

    private void AppendMessage(ImapMailboxNode node)
    {
      var mailbox = node.Mailbox;

      using (var openFileDialog = new OpenFileDialog()) {
        openFileDialog.Filter = fileDialogMessageFilterString;
        openFileDialog.Title = string.Format("{0}へファイルをアップロード", mailbox.Name);
        openFileDialog.Multiselect = false;

        if (openFileDialog.ShowDialog(this) != DialogResult.OK)
          return;

        if (PathUtils.AreExtensionEqual(openFileDialog.FileName, ".eml")) {
          using (var stream = openFileDialog.OpenFile()) {
            mailbox.AppendMessage(stream.Length, DateTimeOffset.Now, delegate(Stream writeStream) {
              stream.CopyTo(writeStream, 0x4000);
            });
          }
        }
        else {
          mailbox.AppendMessage(DateTimeOffset.Now, delegate(Stream writeStream) {
            WriteMimeCapsulatedMessageTo(openFileDialog.FileName, writeStream);
          });
        }

        /* TODO: append-uid
        if (appendedMessageUri != null) {
          var appendedMessage = account.Request(new Uri(appendedMessageUri, string.Format("?UID {0}", ImapStyleUriParser.GetUid(appendedMessageUri))), delegate(ImapWebResponse response) {
            return response.Messages[0];
          });

          listViewMessages.Items.Add(new ImapMessageItem(account, appendedMessage));
        }
        */

        MessageBox.Show(string.Format("{0}を{1}へアップロードしました", openFileDialog.FileName, mailbox.Name));
      }
    }

    private void WriteMimeCapsulatedMessageTo(string path, Stream stream)
    {
      var assemblyName = System.Reflection.Assembly.GetExecutingAssembly().GetName();
      var description = string.Format("{0}\r\nuploaded by {1}", path, assemblyName.Name);

      // TODO: Content-Type: *;format=flowed
      var part1 = new MimeMessage();

      part1.MimeType = MimeType.TextPlain;
      part1.WriteContent(description, ContentTransferEncodingMethod.SevenBit, Encoding.ASCII);

      MimeMessage part2;

      using (var attachmentStream = File.OpenRead(path)) {
        part2 = MimeMessage.CreateAttachment(attachmentStream,
                                             MimeType.GetMimeTypeByExtension(path, mimeTypesFile) ?? MimeType.ApplicationOctetStream,
                                             Path.GetFileName(path),
                                             description,
                                             true /*inline*/);
      }

      var message = MimeMessage.CreateMultipart(new[] {part1, part2});

      message.Headers.Add(new MimeHeader("Date", DateTimeConvert.ToRFC822DateTimeString(DateTimeOffset.Now)));
      message.Headers.Add(new MimeHeader("Subject", path, MimeEncodingMethod.Base64, Encoding.UTF8));
      message.Headers.Add(new MimeHeader("User-Agent", string.Format("{0} {1}", assemblyName.Name, assemblyName.Version)));

      message.Save(stream);
    }

    private void CreateMailbox(TreeNode node)
    {
      ImapAccount account;
      ImapMailboxInfo parentMailbox;

      if (node is ImapMailboxNode) {
        account = (node as ImapMailboxNode).Account;
        parentMailbox = (node as ImapMailboxNode).Mailbox;
      }
      else if (node is ImapAccountNode) {
        account = (node as ImapAccountNode).Account;
        parentMailbox = null;
      }
      else {
        return;
      }

      string createMailboxName;

      using (var dialog = new InputDialog()) {
        dialog.Text = "メールボックスの作成";
        dialog.Description = string.Format("新しく{0}に作成するメールボックスの名前を指定してください。",
                                           (parentMailbox == null) ? account.Name : parentMailbox.Name);
        dialog.InputText = "新しいメールボックス";

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        createMailboxName = dialog.InputText;
      }

      try {
        if (parentMailbox == null)
          account.Client.CreateMailbox(createMailboxName, true);
        else
          parentMailbox.CreateChild(createMailboxName, true);

        AppendStatus("{0}を作成しました", createMailboxName);
      }
      catch (Exception ex) {
        AppendStatus("{0}の作成に失敗しました", createMailboxName);
        AppendStatus(ex);
      }

      UpdateMailboxTree(account);
    }

    private void DeleteMailbox(ImapMailboxNode node)
    {
      var mailbox = node.Mailbox;
      var result = MessageBox.Show(this,
                                   string.Format("{0}を削除します。　この操作は元に戻せません。", mailbox.Name),
                                   "メールボックスの削除", 
                                   MessageBoxButtons.OKCancel,
                                   MessageBoxIcon.Warning,
                                   MessageBoxDefaultButton.Button2);

      if (result != DialogResult.OK)
        return;

      try {
        mailbox.Delete(true);

        AppendStatus("{0}を削除しました", mailbox.Name);
      }
      catch (Exception ex) {
        AppendStatus("{0}の削除に失敗しました", mailbox.Name);
        AppendStatus(ex);
      }

      UpdateMailboxTree(node.Account);
    }

    private void MoveMailboxToTrash(ImapMailboxNode node)
    {
      var account = node.Account;
      var trashMailbox = GetOrCreateTrashMailbox(account);

      if (trashMailbox == null)
        return;

      node.Mailbox.MoveTo(trashMailbox, true);

      UpdateMailboxTree(node.Account);
    }

    private ImapMailboxInfo GetOrCreateTrashMailbox(ImapAccount account)
    {
      foreach (var n in imapMailboxNodeMap.Values) {
        if (n.Account != account)
          continue;

        try {
          return account.Client.GetOrCreateMailbox("Trash", true);
        }
        catch (Exception ex) {
          AppendStatus("メールボックス'Trash'を作成できませんでした");
          AppendStatus(ex);
        }
      }

      return null; // failed
    }

    private void RenameMailbox(ImapMailboxNode node)
    {
      string newMailboxName;

      using (var dialog = new InputDialog()) {
        dialog.Text = "メールボックスの名前の変更";
        dialog.Description = "新しいメールボックスの名前を入力してください。";
        dialog.InputText = node.Mailbox.FullName;

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;
        else if (dialog.InputText == node.Mailbox.FullName)
          return;

        newMailboxName = dialog.InputText;
      }

      try {
        node.Mailbox.MoveTo(newMailboxName, true);

        AppendStatus("{0}の名前を{1}に変更しました", node.Mailbox.Name, newMailboxName);
      }
      catch (Exception ex) {
        AppendStatus("{0}の名前を{1}に変更できませんでした", node.Mailbox.Name, newMailboxName);
        AppendStatus(ex);
      }

      UpdateMailboxTree(node.Account);
    }

    private void ShowSubscribeDialog(TreeNode node)
    {
      ImapAccount account;

      if (node is ImapAccountNode)
        account = (node as ImapAccountNode).Account;
      else if (node is ImapMailboxNode)
        account = (node as ImapMailboxNode).Account;
      else
        return;

      using (var dialog = new SubscribeMailboxDialog(account)) {
        dialog.ShowDialog(this);
      }

      UpdateMailboxTree(account);
    }

    private void UnsubscribeMailbox(ImapMailboxNode node)
    {
      var mailbox = node.Mailbox;

      try {
        mailbox.Unsubscribe();

        AppendStatus("{0}を購読解除しました", mailbox.Name);
      }
      catch (Exception ex) {
        AppendStatus("{0}の購読解除に失敗しました", mailbox.Name);
        AppendStatus(ex);
      }

      UpdateMailboxTree(node.Account);
    }

    private void CancelDelete(PopMailboxNode node)
    {
      try {
        node.Account.Client.CancelDelete();

        node.UpdateText();

        UpdateMessageList(node.Account);

        AppendStatus("{0}の削除をキャンセルしました", node.Account.Name);
      }
      catch (Exception ex) {
        AppendStatus("{0}の削除キャンセルに失敗しました", node.Account.Name);
        AppendStatus(ex);
      }
    }

    private void FetchAndSaveMessage(IMessageContainer message)
    {
      var subject = (message is ImapMessageItem) 
        ? MimeEncoding.Decode((message as ImapMessageItem).Message.Envelope.Subject)
        : GetHeaderValue((message as PopMessageItem).Message, "Subject");

      using (var saveFileDialog = new SaveFileDialog()) {
        saveFileDialog.Filter = fileDialogMessageFilterString;
        saveFileDialog.Title = string.Format("{0}をファイルに保存", message.UniqueId);
        saveFileDialog.FileName = string.Format("{0}.eml", PathUtils.ReplaceInvalidFileNameChars(subject, " "));

        if (saveFileDialog.ShowDialog(this) != DialogResult.OK)
          return;

        using (var stream = saveFileDialog.OpenFile()) {
          using (var responseStream = message.OpenRead(toolStripProgressBarProgress)) {
            responseStream.CopyTo(stream, 0x4000);
          }
        }

        MessageBox.Show(string.Format("{0}を{1}に保存しました", message.UniqueId, saveFileDialog.FileName));
      }
    }

    private void FetchAndSaveSection(SectionDropDownListItem sectionItem)
    {
      if (sectionItem == null)
        return;

      var part = sectionItem.Structure;

      string extension;

      if (part.MediaType.TypeEqualsIgnoreCase("multipart") ||
          part.MediaType.TypeEqualsIgnoreCase("message")) {
        extension = ".eml";
      }
      else if (part.MediaType.TypeEqualsIgnoreCase("text")) {
        if (part.MediaType.SubTypeEqualsIgnoreCase("html"))
          extension = ".html";
        else
          extension = ".txt";
      }
      else {
        extension = string.Empty;
      }

      var disposition = (part is IImapBodyStructureExtension)
        ? (part as IImapBodyStructureExtension).Disposition
        : null;
      var filename = (disposition == null)
        ? null
        : MimeEncoding.DecodeNullable(disposition.Filename);

      if (string.IsNullOrEmpty(filename))
        filename = "section" + part.Section + extension;

      filename = PathUtils.ReplaceInvalidFileNameChars(filename, " ");

      using (var saveFileDialog = new SaveFileDialog()) {
        saveFileDialog.Filter = fileDialogMessageFilterString;
        saveFileDialog.Title = string.Format("{0}をファイルに保存", sectionItem.UniqueId);
        saveFileDialog.FileName = filename;

        if (saveFileDialog.ShowDialog(this) != DialogResult.OK)
          return;

        sectionItem.Message.Save(saveFileDialog.FileName, sectionItem.Structure, ImapMessageFetchBodyOptions.DecodeContent);

        MessageBox.Show(string.Format("{0}を{1}に保存しました", sectionItem.UniqueId, saveFileDialog.FileName));
      }
    }

    private void ToggleMessageSeen(ImapMessageItem item)
    {
      ToggleMessageKeyword(item, ImapMessageFlag.Seen);
    }

    private void ToggleMessageKeywordsDropDownOpening(object sender, EventArgs e)
    {
      var item = sender as ListViewContextMenuItem;

      item.DropDownItems.Clear();

      var mailbox = (treeViewMailbox.SelectedNode as ImapMailboxNode).Account.Client.OpenedMailbox;

      if (mailbox == null)
        return;

      foreach (ImapMessageFlag flag in mailbox.ApplicableFlags) {
        if (flag.IsSystemFlag)
          continue;

        var flagItem = new ListViewTargetFlagContextMenuItem();

        flagItem.Text = flag.Value;
        flagItem.ContextMessage = item.ContextMessage;
        flagItem.TargetFlag = flag;
        flagItem.Checked = (item.ContextMessage as ImapMessageItem).Message.Flags.Contains(flag);
        flagItem.Click += delegate(object clickSender, EventArgs clickEventArgs) {
          var i = clickSender as ListViewTargetFlagContextMenuItem;

          ToggleMessageKeyword(i.ContextMessage as ImapMessageItem, i.TargetFlag);
        };

        item.DropDownItems.Add(flagItem);
      }

      if (!mailbox.IsAllowedToCreateKeywords)
        return;

      toolStripMenuItemMessage_ToggleKeywords_Append.ContextMessage = item.ContextMessage;

      item.DropDownItems.Add(toolStripMenuItemMessage_ToggleKeywords_Append);
    }

    private void StoreMessageKeyword(ImapMessageItem item)
    {
      string keyword = null;

      using (var dialog = new InputDialog()) {
        dialog.Text = "フラグの追加";
        dialog.Description = "新しく設定するフラグの名前を指定してください。";

        /* TODO: validation
        dialog.Validating += delegate(object sender, CancelEventArgs e) {
          e.Cancel = !ImapMessageFlag.IsValidKeyword(dialog.InputText);
          MessageBox.Show("invalid keyword");
        };
        */

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        keyword = dialog.InputText;
      }

      try {
        var storeDataItem = ImapStoreDataItem.AddFlags(keyword);

        item.Message.Store(storeDataItem);

        AppendStatus("{0}に'{1}'を設定しました", item.UniqueId, storeDataItem);
      }
      catch (Exception ex) {
        AppendStatus("{0}に'{1}'を設定できませんでした", item.UniqueId, keyword);
        AppendStatus(ex);
      }
    }

    private void ToggleMessageKeyword(ImapMessageItem item, ImapMessageFlag keyword)
    {
      try {
        item.Message.ToggleFlags(keyword);

        AppendStatus("{0}に'{1}'を設定/設定解除しました", item.UniqueId, keyword);
      }
      catch (Exception ex) {
        AppendStatus("{0}に'{1}'を設定/設定解除できませんでした", item.UniqueId, keyword);
        AppendStatus(ex);
      }
    }

    private void ExpungeMessage(IMessageContainer message)
    {
      try {
        if (message is ImapMessageItem) {
          (message as ImapMessageItem).Message.Delete();
        }
        else {
          var popMessageItem = message as PopMessageItem;

          popMessageItem.MessageInfo.MarkAsDeleted();

          popMessageItem.UpdateView();

          popMailboxNodeMap[popMessageItem.Account].UpdateText();
        }

        AppendStatus("{0}を削除しました", message.UniqueId);
      }
      catch (Exception ex) {
        AppendStatus("{0}を削除できませんでした", message.UniqueId);
        AppendStatus(ex);
      }
    }

    private void CopyOrMoveMessageToDropDownOpening(object sender, EventArgs e)
    {
      var listViewContextMenuItem = sender as ListViewContextMenuItem;

      listViewContextMenuItem.DropDownItems.Clear();

      var messageItem = listViewContextMenuItem.ContextMessage as ImapMessageItem;

      if (messageItem == null)
        return;

      var copy = (sender == toolStripMenuItemMessage_CopyTo);

      foreach (var pair in imapMailboxNodeMap) {
        if (pair.Value.Account == messageItem.Account) {
          var dropDownItem = new ListViewTargetMailboxContextMenuItem();

          dropDownItem.Text = pair.Key.FullName;
          dropDownItem.ContextMessage = messageItem;
          dropDownItem.TargetMailbox = pair.Key;

          if (copy)
            dropDownItem.Click += delegate(object clickSender, EventArgs clickEvent) {
              var item = clickSender as ListViewTargetMailboxContextMenuItem;
              var contextMessageItem = item.ContextMessage as ImapMessageItem;

              CopyMessageTo(contextMessageItem, item.TargetMailbox);
            };
          else
            dropDownItem.Click += delegate(object clickSender, EventArgs clickEvent) {
              var item = clickSender as ListViewTargetMailboxContextMenuItem;
              var contextMessageItem = item.ContextMessage as ImapMessageItem;

              MoveMessageTo(contextMessageItem, item.TargetMailbox);
            };

          listViewContextMenuItem.DropDownItems.Add(dropDownItem);
        }
      }
    }

    private void MoveMessageToTrash(ImapMessageItem item)
    {
      var trashMailbox = GetOrCreateTrashMailbox(item.Account);

      if (trashMailbox == null)
        return;

      CopyOrMoveMessageTo(item, trashMailbox, false);
    }

    private void MoveMessageTo(ImapMessageItem item, ImapMailboxInfo mailbox)
    {
      CopyOrMoveMessageTo(item, mailbox, false);
    }

    private void CopyMessageTo(ImapMessageItem item, ImapMailboxInfo mailbox)
    {
      CopyOrMoveMessageTo(item, mailbox, true);
    }

    private void CopyOrMoveMessageTo(ImapMessageItem item, ImapMailboxInfo mailbox, bool copy)
    {
      try {
        if (copy) {
          item.Message.CopyTo(mailbox);
        }
        else {
          item.Message.MoveTo(mailbox);

          listViewMessages.Items.Remove(item);
        }

        AppendStatus("{0}を{1}にコピーしました", item.UniqueId, mailbox.Name);
      }
      catch (Exception ex) {
        AppendStatus("{0}を{1}にコピーできませんでした", item.UniqueId, mailbox.Name);
        AppendStatus(ex);
        return;
      }
    }

    private void NoOp()
    {
      foreach (var account in imapAccounts) {
        account.Client.Refresh();
      }

      foreach (var account in popAccounts) {
        account.Client.KeepAlive();
      }
    }

    private void LogoutAll()
    {
      foreach (var account in imapAccounts) {
        try {
          account.Client.Logout();
        }
        catch (Exception ex) {
          AppendStatus("{0}からログアウト出来ませんでした", account.ServerUri);
          AppendStatus("Exception: {0}", ex);
        }
      }

      foreach (var account in popAccounts) {
        try {
          account.Client.Logout();
        }
        catch (Exception ex) {
          AppendStatus("{0}からログアウト出来ませんでした", account.ServerUri);
          AppendStatus("Exception: {0}", ex);
        }
      }
    }

    private void AppendStatus(Exception ex)
    {
      AppendStatus("Exception: {0}", ex);
    }

    private void AppendStatus(string format, params object[] args)
    {
      var status = string.Format(format, args);

      toolStripStatusLabelStatusText.Text = status;

      statusStringBuffer.AppendLine(status);

      textBoxStatus.Text = statusStringBuffer.ToString();
      textBoxStatus.Select(textBoxStatus.TextLength - 1, 0);
      textBoxStatus.ScrollToCaret();
      textBoxStatus.Update();

      Console.Error.WriteLine(format, args);
    }

    private class TreeViewContextMenuItem : ToolStripMenuItem {
      public TreeNode ContextNode {
        get; set;
      }
    }

    private class ListViewContextMenuItem : ToolStripMenuItem {
      public IMessageContainer ContextMessage {
        get; set;
      }
    }

    private class ListViewTargetMailboxContextMenuItem : ListViewContextMenuItem {
      public ImapMailboxInfo TargetMailbox {
        get; set;
      }
    }

    private class ListViewTargetFlagContextMenuItem : ListViewContextMenuItem {
      public ImapMessageFlag TargetFlag {
        get; set;
      }
    }

    private class MessageListViewItemComparer : System.Collections.IComparer {
      public SortOrder SortOrder {
        get; private set;
      }

      public int Column {
        get; private set;
      }

      public MessageListViewItemComparer(MessageListViewItemComparer oldComparer, int column)
      {
        if (oldComparer != null && oldComparer.Column == column && oldComparer.SortOrder == SortOrder.Ascending)
          this.SortOrder = SortOrder.Descending;
        else
          this.SortOrder = SortOrder.Ascending;

        this.Column = column;
      }

      public int Compare(object x, object y)
      {
        var xx = (x as ListViewItem).SubItems[Column];
        var yy = (y as ListViewItem).SubItems[Column];

        int sign;

        switch (SortOrder) {
          case SortOrder.Ascending: sign = 1; break;
          case SortOrder.Descending: sign = -1; break;
          default: return 0;
        }

        var comp = (x is IMessageListViewItem)
          ? (x as IMessageListViewItem).CompareTo(Column, yy)
          : string.Compare(xx.Text, yy.Text);

        return sign * comp;
      }
    }

    private interface IAccountContainer {
      AccountBase Account { get; }
    }

    private class ImapAccountNode : TreeNode, IAccountContainer {
      public ImapAccount Account{
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public ImapAccountNode(ImapAccount account)
      {
        this.Account = account;

        Text = string.Format("{0} - {1}", account.Name, account.ServerUri.GetLeftPart(UriPartial.Authority));
      }
    }

    private class PopAccountNode : TreeNode, IAccountContainer {
      public PopAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public PopAccountNode(PopAccount account)
      {
        this.Account = account;

        Text = string.Format("{0} - {1}", account.Name, account.ServerUri.GetLeftPart(UriPartial.Authority));
      }
    }

    private interface IMailboxContainer : IAccountContainer {
      object GetMailbox();
      string GetStatusText();
    }

    private class ImapNonExistentMailboxNode : TreeNode {
      public ImapNonExistentMailboxNode(string text, string mailboxName)
      {
        this.Text = text;
        this.Name = mailboxName;
        this.ForeColor = SystemColors.GrayText;
      }
    }

    private class ImapMailboxNode : TreeNode, IMailboxContainer {
      public ImapAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public ImapMailboxInfo Mailbox {
        get; set;
      }

      public ImapMailboxNode(ImapAccount account, ImapMailboxInfo mailbox)
      {
        this.Account = account;
        this.Mailbox = mailbox;

        this.Name = mailbox.FullName;

        UpdateText();
      }

      public void UpdateText()
      {
        if (!Mailbox.Exists || Mailbox.IsUnselectable)
          ForeColor = SystemColors.GrayText;
        else if (0L < Mailbox.RecentMessageCount)
          ForeColor = Color.Blue;
        else
          ForeColor = Color.Black;

        Text = string.Format("{0} 新着{1} 未読{2} 全{3}",
                             Mailbox.Name,
                             Mailbox.RecentMessageCount,
                             Mailbox.UnseenMessageCount,
                             Mailbox.ExistMessageCount);
      }

      object IMailboxContainer.GetMailbox()
      {
        return Mailbox;
      }

      string IMailboxContainer.GetStatusText()
      {
        return string.Format("'{0}' 全{1}件 (新着{2}件 未読{3}件)",
                             Mailbox.FullName,
                             Mailbox.ExistMessageCount,
                             Mailbox.RecentMessageCount,
                             Mailbox.UnseenMessageCount);
      }
    }

    private class PopMailboxNode : TreeNode, IMailboxContainer {
      public PopAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public PopMailboxNode(PopAccount account)
      {
        this.Account = account;

        UpdateText();
      }

      public void UpdateText()
      {
        Text = (this as IMailboxContainer).GetStatusText();
      }

      object IMailboxContainer.GetMailbox()
      {
        return null;
      }

      string IMailboxContainer.GetStatusText()
      {
        return string.Format(SIPrefixFormat.CurrentInfo, "全{0}件 ({1:F})", Account.Client.MessageCount, Account.Client.TotalSize);
      }
    }

    private interface IMessageContainer : IAccountContainer {
      object Message { get; }
      string UniqueId { get; }
      bool IsDeleted { get; }

      Stream OpenRead(ToolStripProgressBar progressBar);

      TOutput ReadAs<TOutput>(Converter<Stream, TOutput> converter);
    }

    private interface IMessageListViewItem {
      int CompareTo(int column, ListViewItem.ListViewSubItem other);
      void UpdateView();
    }

    private class ImapMessageItem : ListViewItem, IMessageContainer, IMessageListViewItem {
      public ImapAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public ImapMessageInfo Message {
        get; private set;
      }

      public bool IsDeleted {
        get { return Message.IsDeleted; }
      }

      object IMessageContainer.Message {
        get { return Message; }
      }

      public string UniqueId {
        get { return Message.Uid.ToString(); }
      }

      public ImapMessageItem(ImapAccount account, ImapMessageInfo message)
      {
        this.Account = account;
        this.Message = message;

        UpdateView();
      }

      public Stream OpenRead(ToolStripProgressBar progressBar)
      {
        return new NotifyProgressStream(Message.OpenRead(), progressBar);
      }

      public TOutput ReadAs<TOutput>(Converter<Stream, TOutput> converter)
      {
        return Message.ReadAs(converter);
      }

      public void UpdateView()
      {
        this.SubItems.Clear();

        var mediaType = (Message.BodyStructure == null)
          ? null
          : Message.BodyStructure.MediaType;

        this.Text = Message.EnvelopeSubject;
        this.SubItems.Add(Message.Uid.ToString());
        this.SubItems.Add(Message.Length.ToString());
        this.SubItems.Add(mediaType == null ? string.Empty : mediaType.ToString());
        this.SubItems.Add(ToString(Message.InternalDate));
        this.SubItems.Add(ToString(Message.EnvelopeDate));
        this.SubItems.Add(ToString(Message.Envelope.From));
        this.SubItems.Add(Message.Flags.ToString());

        if (!Message.IsDeleted && Message.IsRecent)
          ForeColor = Color.Blue;
        else
          ForeColor = Color.Black;

        if (Message.IsDeleted)
          Font = Fonts.StrikeoutFont;
        else if (Message.IsSeen)
          Font = Control.DefaultFont;
        else
          Font = Fonts.BoldFont;
      }

      public int CompareTo(int column, ListViewItem.ListViewSubItem other)
      {
        switch (column) {
          case 1: return (int)(Message.Uid - long.Parse(other.Text));
          case 2: return (int)(Message.Length - long.Parse(other.Text));
          default: return string.Compare(SubItems[column].Text, other.Text);
        }
      }

      private static string ToString(DateTimeOffset? dto)
      {
        if (dto == null)
          return "-";
        else
          return dto.Value.LocalDateTime.ToString("yyyy-MM-dd HH:mm:ss");
      }

      private static string ToString(ImapAddress[] addresses)
      {
        if (addresses == null || addresses.Length == 0)
          return string.Empty;

        return string.Join(", ", Array.ConvertAll(addresses, delegate(ImapAddress address) {
          return string.Format("{0} <{1}@{2}>",
                               MimeEncoding.DecodeNullable(address.Name),
                               address.Mailbox,
                               address.Host);
        }));
      }
    }

    private class PopMessageItem : ListViewItem, IMessageContainer, IMessageListViewItem {
      public PopAccount Account {
        get; private set;
      }

      AccountBase IAccountContainer.Account {
        get { return Account; }
      }

      public PopMessageInfo MessageInfo {
        get; private set;
      }

      public MimeMessage Message {
        get; private set;
      }

      public bool IsDeleted {
        get { return MessageInfo.IsMarkedAsDeleted; }
      }

      object IMessageContainer.Message {
        get { return MessageInfo; }
      }

      public string UniqueId {
        get { return MessageInfo.UniqueId ?? string.Format("#{0}", MessageInfo.MessageNumber); }
      }

      public PopMessageItem(PopAccount account, PopMessageInfo messageInfo, MimeMessage message)
      {
        this.Account = account;
        this.MessageInfo = messageInfo;

        UpdateMessage(message);
      }

      public void UpdateMessage(MimeMessage message)
      {
        this.Message = message;

        this.Text = MainWindow.GetHeaderValue(message, "Subject");
        this.SubItems.Add(MessageInfo.UniqueId ?? "(Unique-IDを取得できません)");
        this.SubItems.Add(MessageInfo.Length.ToString());
        this.SubItems.Add(MainWindow.GetHeaderValue(message, "Content-Type"));
        this.SubItems.Add("-");
        this.SubItems.Add(MainWindow.GetHeaderValue(message, "Date"));
        this.SubItems.Add(MainWindow.GetHeaderValue(message, "From"));
        this.SubItems.Add("-");
      }

      public void UpdateView()
      {
        if (MessageInfo.IsMarkedAsDeleted)
          Font = Fonts.StrikeoutFont;
        else
          Font = Control.DefaultFont;
      }

      public int CompareTo(int column, ListViewItem.ListViewSubItem other)
      {
        switch (column) {
          case 2: return (int)(MessageInfo.Length - long.Parse(other.Text));
          default: return string.Compare(SubItems[column].Text, other.Text);
        }
      }

      public Stream OpenRead(ToolStripProgressBar progressBar)
      {
        return new NotifyProgressStream(MessageInfo.OpenRead(), progressBar);
      }

      public TOutput ReadAs<TOutput>(Converter<Stream, TOutput> converter)
      {
        return MessageInfo.ReadAs(converter);
      }
    }

    private static string GetHeaderValue(MimeMessage message, string headerName)
    {
      var header = message.Headers.GetHeaders(headerName).FirstOrDefault();

      if (header == null)
        return string.Empty;
      else
        return header.Value;
    }

    private class SectionDropDownList : FormattingDropDownList<SectionDropDownListItem> {
      public SectionDropDownList()
      {
        StringFormatter = SectionDropDownListItem.Format;
      }
    }

    private class SectionDropDownListItem : IMessageContainer {
      public AccountBase Account {
        get; private set;
      }

      public ImapMessageInfo Message {
        get; private set;
      }

      object IMessageContainer.Message {
        get { return Message; }
      }

      public bool IsDeleted {
        get { return Message.IsDeleted; }
      }

      public string UniqueId {
        get { return Message.Uid.ToString(); }
      }

      public IImapBodyStructure Structure {
        get; private set;
      }

      public SectionDropDownListItem(AccountBase account, ImapMessageInfo message, IImapBodyStructure sectionStructure)
      {
        this.Account = account;
        this.Message = message;
        this.Structure = sectionStructure;
      }

      public Stream OpenRead(ToolStripProgressBar progressBar)
      {
        return new NotifyProgressStream(Message.OpenRead(), progressBar);
      }

      public TOutput ReadAs<TOutput>(Converter<Stream, TOutput> converter)
      {
        return Message.ReadAs(converter);
      }

      public static string Format(SectionDropDownListItem item)
      {
        var section = item.Structure.Section;

        return string.Format("{0} ({1}) <{2}>",
                             section.Length == 0 ? "body" : "section " + section,
                             item.Structure.MediaType,
                             item.Structure.Url.LocalPath);
      }
    }

    private class QuotaResourceToolStripItem : ToolStripItem {
      public string QuotaRoot {
        get { return quotaRoot; }
        set
        {
          if (quotaRoot != value) {
            quotaRoot = value;
            OnChanged();
          }
        }
      }

      public ImapQuotaResource Resource {
        get { return resource; }
        set
        {
          if (resource != value) {
            resource = value;
            OnChanged();
          }
        }
      }

      private double UsagePercent {
        get
        {
          if (resource == null || resource.Limit == 0L)
            return 0.0;
          else
            return resource.Usage / (double)resource.Limit;
        }
      }

      public QuotaResourceToolStripItem()
      {
        AutoSize = true;
        AutoToolTip = true;
      }

      private void OnChanged()
      {
        if (resource == null) {
          Text =
          ToolTipText = string.Empty;
        }
        else {
          Text = string.Format("{0}: {1:P1}", resource.Name, UsagePercent);

          if (string.Equals(resource.Name, "STORAGE", StringComparison.OrdinalIgnoreCase))
            ToolTipText = string.Format(SIPrefixFormat.CurrentInfo, "{0:f}/{1:f}", resource.Usage * 1024L, resource.Limit * 1024L);
          else
            ToolTipText = string.Format("{0}/{1}", resource.Usage, resource.Limit);

          ToolTipText = string.Format("[{0}] {1} ({2})", quotaRoot, Text, ToolTipText);
        }

        Invalidate();
      }

      private double GetUsage()
      {
        return (resource.Limit == 0L) ? 0.0 : resource.Usage / (double)resource.Limit;
      }

      protected override void OnPaint(PaintEventArgs e)
      {
        base.OnPaint(e);

        if (resource == null)
          return;

        var g = e.Graphics;
        var bounds = Bounds;

        bounds.X = 0;

        g.FillRectangle(Brushes.DarkGray, bounds);

        using (var b = new LinearGradientBrush(new Point(bounds.Left, bounds.Top),
                                               new Point(bounds.Right, bounds.Top),
                                               Color.Green,
                                               Color.Red)) {
          var blend = new ColorBlend();

          blend.Colors = new Color[] {
            Color.Green,
            Color.Yellow,
            Color.Red,
          };
          blend.Positions = new float[] {
            0.0f,
            0.5f,
            1.0f,
          };

          b.InterpolationColors = blend;

          var fillBounds = bounds;

          fillBounds.Width = (int)(fillBounds.Width * UsagePercent);

          g.FillRectangle(b, fillBounds);
        }

        using (var sf = StringFormat.GenericDefault.Clone() as StringFormat) {
          sf.Alignment = StringAlignment.Center;
          sf.LineAlignment = StringAlignment.Center;

          g.DrawString(Text, Font, Brushes.Black, bounds, sf);
        }

        g.DrawRectangle(Pens.Black, bounds);
      }

      private string quotaRoot;
      private ImapQuotaResource resource;
    }

    private class NotifyProgressStream : InterruptStream {
      public NotifyProgressStream(Stream stream, ToolStripProgressBar progressBar)
        : base(stream)
      {
        this.progressBar = progressBar;

        if (stream.Length == 0L) {
          progressBar.Maximum = int.MaxValue;
          progressBar.Minimum = 0;
          progressBar.Value = 0;
          progressBar.Style = ProgressBarStyle.Continuous;
        }
        else {
          progressBar.Maximum = (int)stream.Length;
          progressBar.Minimum = 0;
          progressBar.Value = 0;
          progressBar.Style = ProgressBarStyle.Blocks;
        }
      }

      public override void Close()
      {
        progressBar.Value = progressBar.Maximum;
        progressBar.Style = ProgressBarStyle.Blocks;

        base.Close();
      }

      protected override int OnRead(byte[] dest, int offset, int count)
      {
        var ret = base.OnRead(dest, offset, count);

        if (100 <= (DateTime.Now - lastUpdated).TotalMilliseconds) {
          progressBar.Value = (int)Position;
          lastUpdated = DateTime.Now;
        }

        return ret;
      }

      private ToolStripProgressBar progressBar;
      private DateTime lastUpdated = DateTime.MinValue;
    }

    private string mimeTypesFile;
    private bool verifyServerCertificates;

    private ImapAccount[] imapAccounts;
    private PopAccount[] popAccounts;
    private AccountBase[] allAccounts;
    private readonly Dictionary<ImapMailboxInfo, ImapMailboxNode> imapMailboxNodeMap = new Dictionary<ImapMailboxInfo, ImapMailboxNode>();
    private readonly Dictionary<PopAccount, PopMailboxNode> popMailboxNodeMap = new Dictionary<PopAccount, PopMailboxNode>();
    private StringBuilder statusStringBuffer = new StringBuilder(10240);
    private ToolStripSeparator contextMenuDummyItem = new ToolStripSeparator();
  }
}
