using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

using Smdn.Net.Imap4;
using Smdn.Net.Imap4.Client;

namespace Smdn.Applications.SimpleMailer {
  public partial class SubscribeMailboxDialog : Form {
    private class MailboxTreeNode : TreeNode {
      public ImapMailboxInfo Mailbox {
        get; private set;
      }

      public MailboxTreeNode(ImapMailboxInfo mailbox)
      {
        this.Mailbox = mailbox;
      }
    }

    public ImapAccount Account {
      get; private set;
    }

    public SubscribeMailboxDialog(ImapAccount account)
    {
      this.Account = account;

      InitializeComponent();

      buttonOK.Click += delegate { Close(); };

      treeViewMailboxes.CheckBoxes = true;
      treeViewMailboxes.NodeMouseClick += delegate(object sender, TreeNodeMouseClickEventArgs e) {
        var node = e.Node as MailboxTreeNode;

        if (node == null)
          return;

        var hitTestInfo = node.TreeView.HitTest(e.Location);

        if (hitTestInfo.Location == TreeViewHitTestLocations.Label)
          node.Checked = !node.Checked;

        SubscribeMailbox(node);
      };
    }

    protected override void OnShown(EventArgs e)
    {
      CreateMailboxNodes();

      base.OnShown(e);
    }

    private void CreateMailboxNodes()
    {
      try {
        Cursor = Cursors.WaitCursor;

        treeViewMailboxes.BeginUpdate();

        treeViewMailboxes.Nodes.Clear();

        try {
          // TODO: mailbox hierarchy
          var nodeMap = new Dictionary<string, MailboxTreeNode>();

          foreach (var listedMailbox in Account.Client.GetMailboxes(ImapMailboxListOptions.Default)) {
            var node = new MailboxTreeNode(listedMailbox);

            node.Checked = false;
            node.Text = listedMailbox.FullName;

            treeViewMailboxes.Nodes.Add(node);

            nodeMap.Add(listedMailbox.FullName, node);
          }

          foreach (var subscribingMailbox in Account.Client.GetMailboxes(ImapMailboxListOptions.SubscribedOnly)) {
            MailboxTreeNode node;

            if (nodeMap.TryGetValue(subscribingMailbox.FullName, out node)) {
              node.Checked = true;
            }
            else {
              node = new MailboxTreeNode(subscribingMailbox);
              node.Checked = true;
              node.Text = subscribingMailbox.FullName;

              treeViewMailboxes.Nodes.Add(node);
            }
          }
        }
        catch (Exception ex) {
          MessageBox.Show(this, string.Format("メールボックスの一覧取得に失敗しました。{0}{1}", Environment.NewLine, ex), "メールボックスの購読", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }
      }
      finally {
        treeViewMailboxes.EndUpdate();
        Cursor = Cursors.Default;
      }
    }

    private void SubscribeMailbox(MailboxTreeNode node)
    {
      try {
        Cursor = Cursors.WaitCursor;

        try {
          if (node.Checked)
            node.Mailbox.Subscribe();
          else
            node.Mailbox.Unsubscribe();
        }
        catch (Exception ex) {
          MessageBox.Show(this, string.Format("リクエストに失敗しました。{0}{1}", Environment.NewLine, ex), "メールボックスの購読", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }
      }
      finally {
        Cursor = Cursors.Default;
      }
    }
  }
}
